# ORF Finder
# Michael Goldwasser

import sys
import re

# let's set some configuration parameters
DEFAULT_FILE = 'MG833025.1.fasta'
DEFAULT_MINLENGTH = 200
DEFAULT_UPSTREAM = 50
CIRCULAR = False


if len(sys.argv) > 1:
    filename = sys.argv[1]
else:
    filename = input('FASTA filename [default: %s]) ' % DEFAULT_FILE).strip()
    if not filename:
        filename = DEFAULT_FILE

if len(sys.argv) > 2:
    minlength = int(sys.argv[2])
else:
    temp = input('Minimum ORF length [default: %d]) ' % DEFAULT_MINLENGTH).strip()
    if not temp.isnumeric():
        minlength = DEFAULT_MINLENGTH
    else:
        minlength = int(temp)

if len(sys.argv) > 3:
    upstream_length = int(sys.argv[3])
else:
    temp = input('Upstream window size [default: %d]) ' % DEFAULT_UPSTREAM).strip()
    if not temp.isnumeric():
        upstream_length = DEFAULT_UPSTREAM
    else:
        upstream_length = int(temp)



START_CODON = 'ATG'
STOP_CODONS = ('TAA', 'TGA', 'TAG')
dnaComp = {'G':'C', 'C':'G', 'A':'T', 'T':'A'}

def reverse_complement(dna):
    """Return the reverse complement strand for the given dna strand."""
    other = ''              # start with an empty string
    for base in dna:
        other += dnaComp.get(base,base)   # if not found, keep base unchanged
    return other[ : :-1]    # reverse the result

def parse_fasta(filename):
    with open(filename) as fasta:
        header = fasta.readline()   # ignore
        return ''.join(line.strip() for line in fasta)


def find_ORF_primary(dna):
    """Return list of ORFs on primary strand, using NCBI (start,stop) values."""
    n = len(dna)
    if CIRCULAR:
        dna = 2*dna
        
    orfs = []
    for frame in (0,1,2):   # restart process with each of the three frames
        k = frame
        active = False
        while k+2 < len(dna):
            codon = dna[k:k+3]
            if codon == START_CODON and k < n and not active:
                j = k             # record starting location
                active = True
            elif active and codon in STOP_CODONS:
                active = False
                if k - j + 1 >= minlength:
                    orfs.append( (j+1,k+3) )     # NCBI style numbering
            k += 3
    return orfs
                    
def find_ORF_secondary(dna):
    """Return list of ORFs on primary strand, using NCBI (start,stop) values."""
    n = len(dna)
    orfs = []
    temp = find_ORF_primary(reverse_complement(dna))
    for (a,b) in temp:
        orfs.append( (n+1-a, n+1-b) )
    return orfs
                    
def find_ORFs(dna):
    """Return a list of all valid ORFs for the dna."""
    orfs = find_ORF_primary(dna) + find_ORF_secondary(dna)
    return orfs

def orf_length(orf):
    """Return the ORF length measured in nucleotides"""
    return 1+abs(orf[0]-orf[1])

def orf_is_primary(orf):
    """Return True if orf is on the primary strand,
       or False if on the complementary strand."""
    return orf[0] < orf[1]

def orf_ref_frame(dna, orf):
    """Return the reference frame containing the orf
       (either 1, 2, or 3, as per NCBI style)."""
    if orf_is_primary(orf):
        return 1 + ((orf[0]+2) % 3)
    else:
        return 1 + ( (len(dna)-orf[0]) % 3)

def orf_sequence(dna, orf):
    """Return the sequence of nucleotides for this orf."""
    if orf_is_primary(orf):
        return dna[orf[0]-1:orf[1]]
    else:
        raw = dna[orf[1]-1:orf[0]]
        return reverse_complement(raw)

def upstream(dna, orf, k):
    """Return the k adjacent nucleotides on the 5' end of the given orf.

    The result is ordered starting from most upstream to least upstream.
    """
    if orf_is_primary(orf):
        start = max(0,orf[0]-k-1)
        return dna[start:orf[0]-1]
    else:
        raw = dna[orf[0]:orf[0]+k]
        return reverse_complement(raw)

def downstream(dna, orf, k):
    """Return the k adjacent nucleotides on the 3' end of the given orf.

    The result is ordered starting from least downstream to most downstream.
    """
    if orf_is_primary(orf):
        return dna[orf[1]:orf[1]+k]
    else:
        raw = dna[orf[1]-k-1:orf[1]-1]
        return reverse_complement(raw)

def format_orf(dna, orf):
    """Return a string description of the orf, akin to NCBI ORF Finder data."""
    strand = '+' if orf_is_primary(orf) else '-'
    frame = orf_ref_frame(dna, orf)
    length = orf_length(orf)
    return '  %s      %d   %7d %7d  %5d|%d' % (
        strand, frame, orf[0], orf[1], length, length//3 - 1)


#------------------------------------------------------------
# Here is where the top-level work begins...
#------------------------------------------------------------
#import promoter

dna = parse_fasta(filename)
orfs = [orf for orf in find_ORFs(dna) if orf_length(orf) >= minlength]
orfs.sort(key=orf_length, reverse=True)

genes = set()
genefile = filename.rsplit('.',1)[0] + '.genes'

try:
    with open(genefile) as gf:

        response = input('Would you like to see identified genes? [y/n] ').strip().lower() == 'y'
        
        if response:
            print('Reading gene file...')
            pattern = re.compile(r'\((\d+),\s*(\d+)\)')
            for line in gf:
                result = pattern.match(line.strip())
                if result:
                    start = int(result.group(1))
                    stop = int(result.group(2))
                    genes.add( (start,stop) )
except IOError:
    pass


uplabelA = ''.join((str(k//10%10) if k >= 10 else ' ') for k in range(1,1+upstream_length))[::-1]
uplabelB = ''.join(str(k%10) for k in range(1,1+upstream_length))[::-1]

print('%d ORFs found' %  len(orfs))

print('                                         '+uplabelA)
print('                                         '+uplabelB)
print('strand frame start  stop  length(nt|aa)  upstream')
for orf in orfs:
    clean = format_orf(dna,orf)
    up = upstream(dna,orf,upstream_length)
    isGene = '  KNOWN GENE/CDS' if orf in genes else ''
    print('{clean:<40} {up:>{width}}'.format(clean=clean, up=up, width=upstream_length) + isGene)
#    promoter.analyze(upstream(dna,orf,100), orf)
